# -*- coding: utf-8 -*-
# wasp_launcher/apps/config.py
#
# Copyright (C) 2016-2017 the wasp-launcher authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-launcher.
#
# Wasp-launcher is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-launcher is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-launcher.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_launcher.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_launcher.version import __status__

import os

from wasp_general.config import WConfig

from wasp_launcher.core import WSyncApp, WAppsGlobals


class WConfigApp(WSyncApp):
	""" Configuration application

	Configuration is generated by applying multiple files. Files are applied in the following order:
	- default configuration
	- configuration file from WConfigApp.__environment_file_var__ path
	- configuration files that are resides in WConfigApp.__environment_dir_var__ (files are ordered by name)
	"""

	__registry_tag__ = 'com.binblob.wasp-launcher.apps.config'
	""" Task tag
	"""

	__dependency__ = ['com.binblob.wasp-launcher.apps.log']
	""" Task dependency
	"""

	__environment_file_var__ = 'WASP_LAUNCHER_CONFIG_FILE'
	""" Environment variable name that is used for configuration filename which overrides defaults
	"""

	__environment_dir_var__ = 'WASP_LAUNCHER_CONFIG_DIR'
	""" Environment variable name that is used for directory where configuration files are resides 
	"""

	__configuration_default__ = os.path.join(os.path.dirname(__file__), '..', 'config', 'defaults.ini')
	""" Environment variable name that is used for configuration filename which overrides defaults.
	"""

	def start(self):
		""" Start this task (all the work is done in :meth:`.WConfigApp.read_config` method)

		:return: None
		"""
		self.read_config()
		WAppsGlobals.log.info('Configuration loading finished')

	def stop(self):
		""" Remove configuration from application.
		Makes :attr:`wasp_launcher.apps.WAppsGlobals.config` configuration unavailable

		:return: None
		"""
		WAppsGlobals.config = None

	@classmethod
	def read_config(cls):
		""" Setup :attr:`wasp_launcher.apps.WAppsGlobals.log` configuration. Reads defaults and
		override it by a file given via :attr:`WConfigApp.__environment_file_var__` environment variable.
		After that configuration files are applied from :attr:`WConfigApp.__environment_dir_var__`

		:return: None
		"""
		WAppsGlobals.config = WConfig()

		def load(filename):
			if os.path.isfile(filename) is False:
				raise RuntimeError("Invalid configuration: '%s'" % filename)
			WAppsGlobals.config.merge(filename)
			WAppsGlobals.log.info('Configuration loaded from file: %s' % os.path.abspath(filename))

		load(cls.__configuration_default__)

		if cls.__environment_file_var__ in os.environ:
			WAppsGlobals.log.info('Variable %s was set' % cls.__environment_file_var__)
			load(os.environ[cls.__environment_file_var__])

		if cls.__environment_dir_var__ in os.environ:
			WAppsGlobals.log.info('Variable %s was set' % cls.__environment_dir_var__)
			config_dir = os.environ[cls.__environment_dir_var__]
			if os.path.isdir(config_dir) is False:
				WAppsGlobals.log.error(
					'Invalid configuration directory was specified: "%s"' % config_dir
				)

			entries = list(os.listdir(config_dir))
			entries.sort()
			for entry in entries:
				config_file = os.path.join(config_dir, entry)
				if os.path.isfile(config_file) is False:
					WAppsGlobals.log.error(
						'Invalid configuration "%s" in the directory "%s"' %
						(config_file, config_dir)
					)
					continue

				load(config_file)
